##
# File:    SchemaDefAccess.py
# Author:  J. Westbrook
# Date:    11-Jun-2011
# Version: 0.001 Initial version
#
# Updates:
#       19-Jun-2018 jdw  add hasSchemaObject() and remove references to __baseTables.
#       22-Jun-2018 jdw  change collection attribute specification to dot notation
#        7-Aug-2018 jdw  add methods to return slice parent details
#       14-Aug-2018 jdw  generalize getDocumentKeyAttributeNames() to return list of attributes ...
#        4-Sep-2018 jdw  add methods to return attribute enumerations
#       18-Nov-2018 jdw  extend the content of COLLECTION_DOCUMENT_ATTRIBUTE_NAMES to _INFO and add method
#                        getDocumentKeyAttributeInfo()
#       16-Jan-2019 jdw  add method getDocumentReplaceAttributeNames() to return COLLECTION_DOCUMENT_REPLACE_ATTRIBUTE_NAMES
#        9-Feb-2019 jdw  add method getSliceExtraSchemaIds()
#       11-Mar-2019 jdw  add getSubCategories(), getAttributeSubCategories(), getSubCategoryAggregates(),
#                        getSubCategoryAggregatesUnitCardinality(), getSubCategorySchemaIdList(),
#                        getSubCategoryAttributeIdList()
##
"""
Schema defintion accessors.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import logging
from operator import itemgetter

logger = logging.getLogger(__name__)


class SchemaDefAccess(object):

    """ Schema definition accessors.

    """

    def __init__(self, schemaDef, **kwargs):
        # def __init__(self, databaseName=None, schemaDefDict=None, convertNames=False, versionedDatabaseName=None, documentDefDict=None, verbose=True):
        #
        #
        self.__name = schemaDef["NAME"] if "NAME" in schemaDef else "unassigned"
        self.__appName = schemaDef["APP_NAME"] if "APP_NAME" in schemaDef else "unassigned"

        self.__databaseName = schemaDef["DATABASE_NAME"] if "DATABASE_NAME" in schemaDef else "unassigned"
        self.__databaseVersion = schemaDef["DATABASE_VERSION"] if "DATABASE_VERSION" in schemaDef else "0_0"
        self.__versionedDatabaseName = self.__databaseName + "_" + self.__databaseVersion
        #
        self.__schemaDefDict = schemaDef["SCHEMA_DICT"] if "SCHEMA_DICT" in schemaDef else {}
        self.__documentDefDict = schemaDef["DOCUMENT_DICT"] if "DOCUMENT_DICT" in schemaDef else {}
        #
        self.__selectionFilterDict = schemaDef["SELECTION_FILTERS"] if "SELECTION_FILTERS" in schemaDef else {}
        self.__sliceParentItemD = schemaDef["SLICE_PARENT_ITEMS"]
        self.__sliceParentFilterD = schemaDef["SLICE_PARENT_FILTERS"]
        self.__sliceIndexD, self.__sliceExtraD = self.__makeSliceIndex()
        self.__nameIndex = self.__makeNameIndex()
        self.__kwargs = kwargs
        #

    def getName(self):
        return self.__name

    def getAppName(self):
        return self.__appName

    def getContentClasses(self, schemaId):
        try:
            return self.__schemaDefDict[schemaId]["CONTENT_CLASSES"]
        except Exception:
            return []

    def getCollectionInfo(self):
        """ Return the collection info [{'NAME': , 'VERSION': xx }, ... ] for the input content type.
        """
        cdL = []
        try:
            cdL = self.__documentDefDict["CONTENT_TYPE_COLLECTION_INFO"]
        except Exception as e:
            logger.error("Failing for with %s", str(e))
        return cdL

    def getCollectionVersion(self, collectionName):
        """ Return the collection info [{'NAME': , 'VERSION': xx }, ... ] for the input content type.
        """
        ret = None
        try:
            for cd in self.__documentDefDict["CONTENT_TYPE_COLLECTION_INFO"]:
                if collectionName == cd["NAME"]:
                    return cd["VERSION"]
        except Exception as e:
            logger.error("Failing for with %s", str(e))
        return ret

    def getVersionedCollection(self, prefix):
        try:
            colL = list(self.__documentDefDict["COLLECTION_DOCUMENT_ATTRIBUTE_ID"].keys())
            for col in colL:
                if col.startswith(prefix):
                    return col
        except Exception as e:
            logger.exception("Faling with %s", str(e))
        return None

    def getCollectionExcluded(self, collectionName):
        """  For input collection, return the list of excluded object ids.

        """
        excludeL = []
        try:
            excludeL = self.__documentDefDict["COLLECTION_CONTENT"][collectionName]["EXCLUDE"]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return excludeL

    def getCollectionSelected(self, collectionName):
        """ For input collection, return the list of selected object ids or the full object id list if no selection is defined.
        """
        sL = []
        try:
            sL = self.__documentDefDict["COLLECTION_CONTENT"][collectionName]["INCLUDE"]
            if sL:
                sL = list(set(sL))
            else:
                sL = self.getSchemaIdList()
            return sL
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))

        return sL

    def getCollectionSliceFilter(self, collectionName):
        """  For input collection, return the defined slice filter

        """
        sf = None
        try:
            sf = self.__documentDefDict["COLLECTION_CONTENT"][collectionName]["SLICE_FILTER"]
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        return sf

    def getCollectionExcludedAttributes(self, collectionName, asSchemaIds=False):
        """  For input collection, return a dictionary of attributes flagged for exclusion.

        """
        eD = None
        try:
            eD = {}
            for catName, atNameL in self.__documentDefDict["COLLECTION_CONTENT"][collectionName]["EXCLUDED_ATTRIBUTES"].items():
                for atName in atNameL:
                    eD[(catName, atName)] = collectionName
            if asSchemaIds:
                # convert to internal schema identifiers if asSchemaIds is set.
                rD = {}
                for (schemaName, atName), cn in eD.items():
                    if schemaName in self.__nameIndex and atName in self.__nameIndex[schemaName]["ATTRIBUTES"]:
                        rD[(self.__nameIndex[schemaName]["SCHEMA_ID"], self.__nameIndex[schemaName]["ATTRIBUTES"][atName])] = cn
                eD = rD
            #
        except Exception as e:
            logger.debug("Collection %s failing with %s", collectionName, str(e))
        #
        logger.debug("collectionName %s excluded attributes %r", collectionName, eD)
        return eD

    def getDataSelectors(self, selectorName):
        sL = []
        if selectorName in self.__selectionFilterDict:
            sL = self.__selectionFilterDict[selectorName]
        return sL

    def getDataSelectorNames(self):
        return list(self.__selectionFilterDict.keys())

    def getDocumentKeyAttributeNames(self, collectionName):
        """ Return list of key document attributes required to uniquely identify a document.
        """
        ret = []
        try:
            return list(self.__documentDefDict["COLLECTION_DOCUMENT_ATTRIBUTE_NAMES"][collectionName])
        except Exception as e:
            logger.exception("Failing for collection %s with %r", collectionName, str(e))
        return ret

    def getDocumentReplaceAttributeNames(self, collectionName):
        """ Return list of document attributes required to remove all relevant documents
            prior  to load 'replace' operation.
        """
        ret = []
        try:
            return list(self.__documentDefDict["COLLECTION_DOCUMENT_REPLACE_ATTRIBUTE_NAMES"][collectionName])
        except Exception as e:
            logger.exception("Failing for collection %s with %r", collectionName, str(e))
        return ret

    def getPrivateDocumentAttributes(self, collectionName):
        """ Return list of key document identifiers and document private document names.
        """
        ret = []
        try:
            return [d for d in self.__documentDefDict["COLLECTION_DOCUMENT_PRIVATE_KEYS"][collectionName]]
        except Exception as e:
            logger.exception("Failing for collection %s with %r", collectionName, str(e))
        return ret

    def getDocumentIndices(self, collectionName):
        """ Return list of document indices.
        """
        ret = []
        try:
            return [d for d in self.__documentDefDict["COLLECTION_DOCUMENT_INDICES"][collectionName]]
        except Exception as e:
            logger.exception("Failing for collection %s with %r", collectionName, str(e))
        return ret

    def getDocumentIndex(self, collectionName, indexName):
        """ Return the attribute list for a particular document index.
        """
        ret = []
        try:
            for dD in self.__documentDefDict["COLLECTION_DOCUMENT_INDICES"][collectionName]:
                if dD["INDEX_NAME"] == indexName:
                    return dD["ATTRIBUTE_NAMES"]
        except Exception as e:
            logger.exception("Failing for collection %s index %r with %r", collectionName, indexName, str(e))
        return ret

    def getSubCategoryAggregates(self, collectionName):
        """  Return the list of subcategory aggregates for the input collection.
        """
        try:
            return [d["NAME"] for d in self.__documentDefDict["COLLECTION_SUB_CATEGORY_AGGREGATES"][collectionName]]
        except Exception:
            # logger.exception('Failing with %s' % str(e))
            pass
        return []

    def getSubCategorySchemaIdList(self, subCategoryName):
        """ Return the schema Ids containing the input subCategory
        """
        sIdL = []
        try:
            for sId in self.__schemaDefDict:
                dD = self.__schemaDefDict[sId]
                if subCategoryName in dD["SCHEMA_SUB_CATEGORIES"]:
                    sIdL.append(sId)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return sIdL

    def getSubCategoryAttributeIdList(self, schemaId, subCategoryName):
        """ Return the schema attribute Ids containing the input subCategory
        """
        atIdL = []
        try:
            tD = self.__schemaDefDict[schemaId]
            for atId, v in tD["ATTRIBUTE_INFO"].items():
                if subCategoryName in v["SUB_CATEGORIES"]:
                    atIdL.append(atId)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return atIdL

    def getSubCategoryAggregatesUnitCardinality(self, collectionName, subCategoryName):
        """  Return if the input subcategory aggregate has unit cardinality.
        """
        try:
            for dD in self.__documentDefDict["COLLECTION_SUB_CATEGORY_AGGREGATES"][collectionName]:
                if subCategoryName == dD["NAME"]:
                    return dD["HAS_UNIT_CARDINALITY"]
        except Exception:
            return False

    def hasUnitCardinality(self, schemaId):
        try:
            return self.__schemaDefDict[schemaId]["SCHEMA_UNIT_CARDINALITY"]
        except Exception:
            return False

    def getDatabaseName(self):
        return self.__databaseName

    def getVersionedDatabaseName(self):
        return self.__versionedDatabaseName

    def hasSchemaObject(self, schemaId):
        return schemaId in self.__schemaDefDict

    def getSchemaObject(self, schemaId):
        return SchemaDef(schemaDefDict=self.__schemaDefDict[schemaId])

    def getSchemaName(self, schemaId):
        try:
            return self.__schemaDefDict[schemaId]["SCHEMA_NAME"]
        except Exception:
            return None

    def getAttributeName(self, schemaId, attributeId):
        try:
            return self.__schemaDefDict[schemaId]["ATTRIBUTES"][attributeId]
        except Exception:
            return None

    def getSchemaIdList(self):
        return list(self.__schemaDefDict.keys())

    def getAttributeIdList(self, schemaId):
        tD = self.__schemaDefDict[schemaId]
        tupL = []
        for attributeId, v in tD["ATTRIBUTE_INFO"].items():
            tupL.append((attributeId, v["ORDER"]))
        sTupL = sorted(tupL, key=itemgetter(1))
        return [tup[0] for tup in sTupL]

    def getDefaultAttributeParameterMap(self, schemaId, skipAuto=True):
        """ For the input table, return a dictionary of attribute identifiers and parameter names.
            Default parameter names are compressed and camel-case conversions of the attribute ids.
        """
        dL = []
        aIdList = self.getAttributeIdList(schemaId)

        try:
            tDef = self.getSchemaObject(schemaId)
            for aId in aIdList:
                if skipAuto and tDef.isAutoIncrementType(aId):
                    continue
                pL = aId.lower().split("_")
                tL = []
                tL.append(pL[0])
                for pV in pL[1:]:
                    tt = pV[0].upper() + pV[1:]
                    tL.append(tt)
                dL.append((aId, "".join(tL)))
        except Exception:
            for aId in aIdList:
                dL.append((aId, aId))
        return dL

    def getAttributeNameList(self, schemaId):
        tD = self.__schemaDefDict[schemaId]
        tupL = []
        for k, v in tD["ATTRIBUTE_INFO"].items():
            attributeName = tD["ATTRIBUTES"][k]
            tupL.append((attributeName, v["ORDER"]))
        sTupL = sorted(tupL, key=itemgetter(1))
        return [tup[0] for tup in sTupL]

    def getQualifiedAttributeName(self, tableAttributeTuple=(None, None)):
        schemaId = tableAttributeTuple[0]
        attributeId = tableAttributeTuple[1]
        tD = self.__schemaDefDict[schemaId]
        tableName = tD["SCHEMA_NAME"]
        attributeName = tD["ATTRIBUTES"][attributeId]
        qAN = tableName + "." + attributeName
        return qAN

    def __makeSliceIndex(self):
        """  Return a dictionary of slice parent and child attribures together with schema objects
             to be included in every slice.

        """
        sliceD = {}
        extraD = {}
        try:
            sliceNames = self.getSliceNames()
            for sliceName in sliceNames:
                dD = {}
                for schemaId, tD in self.__schemaDefDict.items():
                    if sliceName in tD["SLICE_ATTRIBUTES"]:
                        if schemaId not in dD:
                            dD[schemaId] = {}
                        for sD in tD["SLICE_ATTRIBUTES"][sliceName]:
                            # d[schemaId][(sD['PARENT_CATEGORY'], sD['PARENT_ATTRIBUTE'])] = sD['CHILD_ATTRIBUTE']
                            dD[schemaId].setdefault((sD["PARENT_CATEGORY"], sD["PARENT_ATTRIBUTE"]), []).append(sD["CHILD_ATTRIBUTE"])
                    #
                    if sliceName in tD["SLICE_CATEGORY_EXTRAS"] and tD["SLICE_CATEGORY_EXTRAS"][sliceName]:
                        extraD.setdefault(sliceName, []).append(schemaId)

                sliceD[sliceName] = dD
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return sliceD, extraD

    def getSliceNames(self):
        """ Return a list of the slice names defined for the current schema
        """
        try:
            return list(self.__sliceParentItemD.keys())
        except Exception:
            return []

    def getSliceParentItems(self, sliceName):
        """ Return a list of dictionaries containing the parent items for input slice name

        """
        try:
            return self.__sliceParentItemD[sliceName]
        except Exception:
            return []

    def getSliceParentFilters(self, sliceName):
        """ Return a list of dictionaries containing the parent items for input slice name
        """
        try:
            return self.__sliceParentFilterD[sliceName]
        except Exception:
            return []

    def getSliceIndex(self, sliceName):
        try:
            return self.__sliceIndexD[sliceName]
        except Exception:
            return {}

    def getSliceExtraSchemaIds(self, sliceName):
        return self.__sliceExtraD[sliceName] if sliceName in self.__sliceExtraD else []

    def __makeNameIndex(self):
        """Return an index of schema and attribute names resolving to corresponding internal ids.

        Returns:
            dict : {<schemaName> : {'SCHEMA_ID": ... , 'ATTRIBUTES': {<atName>: <atId>, ... }}

        """
        nameIndexD = {}
        for schemaId, sD in self.__schemaDefDict.items():
            schemaName = sD["SCHEMA_NAME"]
            atD = {atName: atId for atId, atName in sD["ATTRIBUTES"].items()}
            nameIndexD[schemaName] = {"SCHEMA_ID": schemaId, "ATTRIBUTES": atD}
        return nameIndexD


class SchemaDef(object):

    """  Wrapper class for table schema definition.
    """

    def __init__(self, schemaDefDict=None):
        self.__tD = schemaDefDict if schemaDefDict else {}
        self.__nomalizedEnumD = {}

    def getName(self):
        try:
            return self.__tD["SCHEMA_NAME"]
        except Exception:
            return None

    def isMandatory(self):
        try:
            return self.__tD["SCHEMA_MANDATORY"]
        except Exception:
            return False

    def getType(self):
        try:
            return self.__tD["SCHEMA_TYPE"]
        except Exception:
            return None

    def getContentClasses(self):
        try:
            return self.__tD["SCHEMA_CONTENT_CLASSES"]
        except Exception:
            return []

    def getSubCategories(self):
        try:
            return self.__tD["SCHEMA_SUB_CATEGORIES"]
        except Exception:
            return []

    def getId(self):
        try:
            return self.__tD["SCHEMA_ID"]
        except Exception:
            return None

    def getAttributeIdDict(self):
        try:
            return self.__tD["ATTRIBUTES"]
        except Exception:
            return {}

    def getAttributeNameDict(self):
        try:
            return {v: k for k, v in self.__tD["ATTRIBUTES"].items()}
        except Exception:
            return {}

    def getAttributeName(self, attributeId):
        try:
            return self.__tD["ATTRIBUTES"][attributeId]
        except Exception:
            return None

    # deprecated
    def getMapAttributeInfo(self, attributeId):
        """ Return the tuple of mapping details for the input attribute id.
        """
        raise NotImplementedError()

    def getAttributeType(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"]
        except Exception:
            return None

    def isAutoIncrementType(self, attributeId):
        try:
            tL = [tt.upper() for tt in self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].split()]
            if "AUTO_INCREMENT" in tL:
                return True
        except Exception:
            pass
        return False

    def isAttributeStringType(self, attributeId):
        try:
            return self.__isStringType(self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].upper())
        except Exception:
            return False

    def isAttributeFloatType(self, attributeId):
        try:
            return self.__isFloatType(self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].upper())
        except Exception:
            return False

    def isAttributeIntegerType(self, attributeId):
        try:
            return self.__isIntegerType(self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].upper())
        except Exception:
            return False

    def isAttributeDateType(self, attributeId):
        try:
            return self.__isDateType(self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].upper())
        except Exception:
            return False

    def getAttributeWidth(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["WIDTH"]
        except Exception:
            return None

    def getAttributePrecision(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["PRECISION"]
        except Exception:
            return None

    def getAttributeNullable(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["NULLABLE"]
        except Exception:
            return None

    def getAttributeIsPrimaryKey(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["PRIMARY_KEY"]
        except Exception:
            return None

    def getAttributeEnumList(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["ENUMERATION"]
        except Exception:
            return []

    def normalizeEnum(self, attributeId, enum):
        try:
            if self.__tD["ATTRIBUTE_INFO"][attributeId]["ENUMERATION"] and attributeId not in self.__nomalizedEnumD:
                self.__nomalizedEnumD[attributeId] = {str(ky).lower(): ky for ky in self.__tD["ATTRIBUTE_INFO"][attributeId]["ENUMERATION"]}
            return self.__nomalizedEnumD[attributeId][str(enum).lower()]
        except Exception as e:
            logger.debug("Failing for %s  %s and %r - %r with %r", self.__tD["SCHEMA_NAME"], attributeId, enum, self.__nomalizedEnumD, str(e))
        return enum

    def isEnumerated(self, attributeId):
        try:
            return len(self.__tD["ATTRIBUTE_INFO"][attributeId]["ENUMERATION"]) > 0
        except Exception:
            return False

    def getAttributeSubCategories(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["SUB_CATEGORIES"]
        except Exception:
            return []

    def getAttributeFilterTypes(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["FILTER_TYPES"]
        except Exception:
            return []

    def getPrimaryKeyAttributeIdList(self):
        try:
            return [atId for atId in self.__tD["ATTRIBUTE_INFO"].keys() if self.__tD["ATTRIBUTE_INFO"][atId]["PRIMARY_KEY"]]
        except Exception:
            pass

        return []

    def getAttributeIdList(self):
        """ Get the ordered attribute Id list
        """
        tupL = []
        for k, v in self.__tD["ATTRIBUTE_INFO"].items():
            tupL.append((k, v["ORDER"]))
        sTupL = sorted(tupL, key=itemgetter(1))
        return [tup[0] for tup in sTupL]

    def getAttributeNameList(self):
        """ Get the ordered attribute name list
        """
        tupL = []
        for k, v in self.__tD["ATTRIBUTE_INFO"].items():
            tupL.append((k, v["ORDER"]))
        sTupL = sorted(tupL, key=itemgetter(1))
        return [self.__tD["ATTRIBUTES"][tup[0]] for tup in sTupL]

    def getIndexNames(self):
        try:
            return list(self.__tD["INDICES"].keys())
        except Exception:
            return []

    def getIndexType(self, indexName):
        try:
            return self.__tD["INDICES"][indexName]["TYPE"]
        except Exception:
            return None

    def getIndexAttributeIdList(self, indexName):
        try:
            return self.__tD["INDICES"][indexName]["ATTRIBUTES"]
        except Exception:
            return []

    def getMapAttributeNameList(self):
        """ Get the ordered mapped attribute name list
        """
        try:
            tupL = []
            for k in self.__tD["ATTRIBUTE_MAP"].keys():
                iOrd = self.__tD["ATTRIBUTE_INFO"][k]["ORDER"]
                tupL.append((k, iOrd))

            sTupL = sorted(tupL, key=itemgetter(1))
            return [self.__tD["ATTRIBUTES"][tup[0]] for tup in sTupL]
        except Exception:
            return []

    def getMapAttributeIdList(self):
        """ Get the ordered mapped attribute name list
        """
        try:
            tupL = []
            for k in self.__tD["ATTRIBUTE_MAP"].keys():
                iOrd = self.__tD["ATTRIBUTE_INFO"][k]["ORDER"]
                tupL.append((k, iOrd))

            sTupL = sorted(tupL, key=itemgetter(1))

            return [tup[0] for tup in sTupL]
        except Exception:
            return []

    def getMapInstanceCategoryList(self):
        """ Get the unique list of data instance categories within the attribute map.
        """
        try:
            cL = [vD["CATEGORY"] for k, vD in self.__tD["ATTRIBUTE_MAP"].items() if vD["CATEGORY"] is not None]  #
            uL = list(set(cL))
            return uL
        except Exception:
            return []

    def getMapOtherAttributeIdList(self):
        """ Get the list of attributes that have no assigned data instance mapping.
        """
        try:
            aL = []
            for k, vD in self.__tD["ATTRIBUTE_MAP"].items():
                if vD["CATEGORY"] is None:
                    aL.append(k)
            return aL
        except Exception:
            return []

    def getMapInstanceAttributeList(self, categoryName):
        """ Get the list of instance category attribute names for mapped attributes in the input instance category.
        """
        try:
            aL = []
            for _, vD in self.__tD["ATTRIBUTE_MAP"].items():
                if vD["CATEGORY"] == categoryName:
                    aL.append(vD["ATTRIBUTE"])
            return aL
        except Exception:
            return []

    def getMapInstanceAttributeIdList(self, categoryName):
        """ Get the list of schema attribute Ids for mapped attributes from the input instance category.
        """
        try:
            aL = []
            for k, vD in self.__tD["ATTRIBUTE_MAP"].items():
                if vD["CATEGORY"] == categoryName:
                    aL.append(k)
            return aL
        except Exception:
            return []

    def getMapAttributeFunction(self, attributeId):
        """ Return the tuple element of mapping details for the input attribute id for the optional function.
        """
        try:
            return self.__tD["ATTRIBUTE_MAP"][attributeId]["METHOD_NAME"]
        except Exception:
            return None

    def getMapAttributeFunctionArgs(self, attributeId):
        """ Return the tuple element of mapping details for the input attribute id for the optional function arguments.
        """
        try:
            return self.__tD["ATTRIBUTE_MAP"][attributeId]["ARGUMENTS"]
        except Exception:
            return None

    def getMapAttributeIdDict(self):
        """ Return the dictionary of d[schema attribute id] = mapped instance category attribute

            Exclude and attributes that lack a schema mapping (e.g. functional mappings )
        """
        dD = {}
        for k, vD in self.__tD["ATTRIBUTE_MAP"].items():
            if vD["ATTRIBUTE"]:
                dD[k] = vD["ATTRIBUTE"]
        return dD

    def getMapAttributeNameDict(self):
        """ Return the dictionary of d[schema attribute id] = mapped instance category attribute

            Exclude and attributes that lack a schema mapping (e.g. functional mappings )
        """
        dD = {}
        for k, vD in self.__tD["ATTRIBUTE_MAP"].items():
            if vD["ATTRIBUTE"]:
                dD[vD["ATTRIBUTE"]] = k
        return dD

    def getMapMergeIndexAttributes(self, categoryName):
        """  Return the list of merging index attribures for this mapped instance category.
        """
        try:
            return self.__tD["MAP_MERGE_INDICES"][categoryName]["ATTRIBUTES"]
        except Exception:
            return []

    def getMapMergeIndexType(self, indexName):
        """ Return the merging index type for this mapped instance category.
        """
        try:
            return self.__tD["MAP_MERGE_INDICES"][indexName]["TYPE"]
        except Exception:
            return []

    def getAppNullValue(self, attributeId):
        """ Return the appropriate NULL value for this attribute:.
        """
        try:
            if self.__isStringType(self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].upper()):
                return ""
            elif self.__isDateType(self.__tD["ATTRIBUTE_INFO"][attributeId]["APP_TYPE"].upper()):
                return r"\N"
            else:
                return r"\N"
        except Exception:
            return r"\N"

    def getAppNullValueDict(self):
        """ Return a dictionary containing appropriate NULL value for each attribute.
        """
        dD = {}
        atId = atInfo = None
        try:
            for atId, atInfo in self.__tD["ATTRIBUTE_INFO"].items():

                if self.__isStringType(atInfo["APP_TYPE"].upper()):
                    dD[atId] = ""
                elif self.__isDateType(atInfo["APP_TYPE"].upper()):
                    dD[atId] = r"\N"
                else:
                    dD[atId] = r"\N"
        except Exception as e:
            logger.exception("Failing with atId %r atInfo %r with %s", atId, atInfo, str(e))
        #
        return dD

    def getStringWidthDict(self):
        """ Return a dictionary containing maximum string widths assigned to char data types.
            Non-character type data items are assigned zero width.
        """
        dD = {}
        for atId, atInfo in self.__tD["ATTRIBUTE_INFO"].items():
            if self.__isStringType(atInfo["APP_TYPE"].upper()):
                dD[atId] = int(atInfo["WIDTH"])
            else:
                dD[atId] = 0
        return dD

    def isIterable(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["ITERABLE_DELIMITER"] is not None
        except Exception:
            return False

    def isOtherAttributeType(self, attributeId):
        """ Get the list of attributes that have no assigned instance mapping.
        """
        try:
            return self.__tD["ATTRIBUTE_MAP"][attributeId]["CATEGORY"] is None
        except Exception:
            return False

    def getIterableSeparator(self, attributeId):
        try:
            return self.__tD["ATTRIBUTE_INFO"][attributeId]["ITERABLE_DELIMITER"]
        except Exception:
            return None

    def __isStringType(self, sqlType):
        """ Return if input type corresponds to a common SQL string data type.
        """
        try:
            return sqlType.upper() in ["VARCHAR", "CHAR", "TEXT", "MEDIUMTEXT", "LONGTEXT"]
        except Exception:
            return False

    def __isDateType(self, sqlType):
        """ Return if input type corresponds to a common SQL date/time data type.
        """
        try:
            return sqlType.upper() in ["DATE", "DATETIME"]
        except Exception:
            return False

    def __isFloatType(self, sqlType):
        """ Return if input type corresponds to a common SQL string data type.
        """
        try:
            return sqlType.upper() in ["FLOAT", "DECIMAL", "DOUBLE PRECISION", "NUMERIC"]
        except Exception:
            return False

    def __isIntegerType(self, sqlType):
        """ Return if input type corresponds to a common SQL string data type.
        """
        try:
            return sqlType.upper().startswith("INT") or sqlType.upper() in ["INTEGER", "BIGINT", "SMALLINT"]
        except Exception:
            return False

    def getDeleteAttributeId(self):
        """ Return the attribute identifier that is used to delete all of the
            records associated with the highest level of organizaiton provided by
            this schema definition (e.g. entry, definition, ...).
        """
        try:
            return self.__tD["SCHEMA_DELETE_ATTRIBUTE"]
        except Exception:
            return None

    def getDeleteAttributeName(self):
        """ Return the attribute name that is used to delete all of the
            records associated with the highest level of organizaiton provided by
            this schema definition (e.g. entry, definition, ...).
        """
        try:
            return self.__tD["ATTRIBUTES"][self.__tD["SCHEMA_DELETE_ATTRIBUTE"]]
        except Exception:
            return None

    def hasSliceAttributes(self, sliceName):
        """ Return True if slice attributes are defined for this schema object.
        """
        try:
            if self.__tD["SLICE_ATTRIBUTES"][sliceName]:
                return True
            else:
                return False
        except Exception:
            return False

    def hasSliceUnitCardinality(self, sliceName):
        """ Return True if slice for this schema object has unit cardinality.
        """
        try:
            return self.__tD["SLICE_UNIT_CARDINALITY"][sliceName]
        except Exception:
            return False

    def isSliceExtra(self, sliceName):
        """ Return True if this schema object is extra included content in the input slice.
        """
        try:
            return self.__tD["SLICE_CATEGORY_EXTRAS"][sliceName]
        except Exception:
            return False

    def getSliceAttributeId(self, sliceName, parentCategoryId, parentAttributeId):
        """ Naive method to return corresponding child attribute for the input slice parent.
        """
        try:
            for dD in self.__tD["SLICE_ATTRIBUTES"][sliceName]:
                if dD["PARENT_CATEGORY"] == parentCategoryId and dD["PARENT_ATTRIBUTE"] == parentAttributeId:
                    return dD["CHILD_ATTRIBUTE"]
            return None
        except Exception:
            return None



##
# File:    DictionaryApiProvider.py
# Author:  J. Westbrook
# Date:    3-Jun-2019
# Version: 0.001 Initial version
#
# Updates:
#  14-Aug-2019 jdw adding remote dictionary fetch and caching logic.
##
"""
Resource provider for dictionary API.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from mmcif.api.DictionaryApi import DictionaryApi
from rcsb.utils.io.FileUtil import FileUtil
from rcsb.utils.io.MarshalUtil import MarshalUtil
from rcsb.utils.io.SingletonClass import SingletonClass

logger = logging.getLogger(__name__)


class DictionaryApiProvider(SingletonClass):
    """ Resource provider for dictionary APIs.
    """

    def __init__(self, dirPath, useCache=True):
        """Resource provider for dictionary APIs.

        Args:
            dirPath (str): path to the directory containing cache files
            useCache (bool, optional): flag to use cached files. Defaults to True.

        """
        self.__apiMap = {}
        self.__dirPath = dirPath
        self.__useCache = useCache
        #
        self.__fileU = FileUtil(workPath=self.__dirPath)
        logger.debug("Leaving constructor")

    def __reload(self, dictLocators, dirPath, useCache=True):
        """Reload local cache of dictionary resources and return a dictionary API instance.

        Args:
            dictLocators (list, str): list of locators for dictionary resource files
            dirPath (str): path to the directory containing cache files
            useCache (bool, optional): flag to use cached files. Defaults to True.

        Returns:
            (object): instance of dictionary API
        """
        #
        # verify the exitence of the cache directory ...
        self.__fileU.mkdir(dirPath)
        if not useCache:
            for dictLocator in dictLocators:
                try:
                    fn = self.__fileU.getFileName(dictLocator)
                    os.remove(os.path.join(dirPath, fn))
                except Exception:
                    pass
        #
        ret = True
        for dictLocator in dictLocators:
            cacheFilePath = os.path.join(dirPath, self.__fileU.getFileName(dictLocator))
            if useCache and self.__fileU.exists(cacheFilePath):
                # nothing to do
                continue
            logger.debug("Fetching url %s caching in %s", dictLocator, cacheFilePath)
            ok = self.__fileU.get(dictLocator, cacheFilePath)
            ret = ret and ok
        return ret

    def getApi(self, dictLocators, **kwargs):
        """Return a dictionary API object of the input dictioaries.

        Arguments:
            dictLocators {list str} -- list of dictionary locator paths

        Returns:
            [object] -- returns DictionaryApi() object for input dictionaries
        """
        dictFileNames = [self.__fileU.getFileName(dictLocator) for dictLocator in dictLocators]
        dictTup = tuple(dictFileNames)
        dApi = self.__apiMap[dictTup] if dictTup in self.__apiMap else self.__getApi(dictLocators, **kwargs)
        self.__apiMap[dictTup] = dApi
        return dApi

    def __getApi(self, dictLocators, **kwargs):
        """ Return an instance of a dictionary API instance for the input dictionary locator list.
        """
        consolidate = kwargs.get("consolidate", True)
        replaceDefinition = kwargs.get("replaceDefinitions", True)
        verbose = kwargs.get("verbose", True)
        #
        ok = self.__reload(dictLocators, self.__dirPath, useCache=self.__useCache)
        #
        dApi = None
        if ok:
            mU = MarshalUtil()
            containerList = []
            for dictLocator in dictLocators:
                cacheFilePath = os.path.join(self.__dirPath, self.__fileU.getFileName(dictLocator))
                containerList.extend(mU.doImport(cacheFilePath, fmt="mmcif-dict"))
            #
            dApi = DictionaryApi(containerList=containerList, consolidate=consolidate, replaceDefinition=replaceDefinition, verbose=verbose)
        return dApi



##
# File: TextUtil.py
# Date: 8-Jan-2019
#
# Collection of text utilities -
#
##

try:
    from html import unescape  # python 3.4+
except ImportError:
    try:
        from html.parser import HTMLParser  # python 3.x (<3.4)
    except ImportError:
        from HTMLParser import HTMLParser  # python 2.x
    unescape = HTMLParser().unescape


def unescapeXmlCharRef(iStr):
    """
    Convert html character entities into unicode.
    """
    try:
        return unescape(iStr)
    except Exception:
        return iStr


if __name__ == "__main__":
    print("BEGIN")
    print("%r" % unescapeXmlCharRef(None))
    print("%r" % unescapeXmlCharRef(""))
    print("%r" % unescapeXmlCharRef("&lt;b&gt;"))
    print("%r" % unescapeXmlCharRef("Here is a &quot;").encode("utf-8"))
    print("%r" % unescapeXmlCharRef("Here is a &Phi;").encode("utf-8"))
    print("%r" % unescapeXmlCharRef("Here is a &Psi;"))
    print("%r" % unescapeXmlCharRef("Here is a &alpha;"))
    print("%r" % unescapeXmlCharRef("Here is a &#xa3;"))

    print("%r" % unescapeXmlCharRef("Here is a &#8453;"))
    print("%r" % unescapeXmlCharRef("Here is a &#9734;"))
    print("%r" % unescapeXmlCharRef("Here is a &#120171;"))
    for ichar in range(1, 8000):
        myStr = "decimal %6d char &#%d;" % (ichar, ichar)
        print("%r" % unescapeXmlCharRef(myStr))
    print("END")



##
# File:    DictMethodRunner.py
# Author:  J. Westbrook
# Date:    18-Aug-2018
# Version: 0.001 Initial version
#
# Updates:
# 12-Nov-2018 jdw Run block methods after category and attribute methods.
#  5-Jun-2019 jdw Refactor and generalize and remove dependencies on rcsb.db package
# 17-Jul-2019 jdw Propagate kwargs to __getModuleInstance()
#
##
"""
Manage the invocation of dictionary methods implemented in helper classes.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import sys
from operator import itemgetter

# from mmcif.api.DictionaryApi import DictionaryApi

logger = logging.getLogger(__name__)


class DictMethodRunner(object):
    """Manage the invocation of dictionary methods implemented as class methods.
    """

    def __init__(self, dictionaryApi, modulePathMap=None, **kwargs):
        """Manage invocation of dictionary methods referenced in external modules.

        Arguments:
            dictionaryApi {object} -- instance of DictionaryApi() for dictionary with target method definitions

        Keyword Arguments:
            modulePathMap {dict str} -- mapping between dictionary module path and execution path (default: {None})
            cacheModuleFlag {bool} -- flag to cache module instances (defaullt: True)
            implentationSource {str} -- method implementation (default: 'reference')
            methodCodes (list str) -- filter methods by codes {default: ['calculation']}
        """
        self.__dApi = dictionaryApi
        self.__modulePathMap = modulePathMap if modulePathMap else {}
        self.__cacheModuleFlag = kwargs.get("cacheModuleFlag", True)
        methodCodes = kwargs.get("methodCodes", ["calculation"])
        implementationSource = kwargs.get("implementationCodes", "reference")
        #
        self.__kwargs = kwargs
        #
        # Preserve and reuse the module instances if caching is enabled
        self.__moduleCache = {}
        #
        self.__methodD = self.__getMethodInfo(implementationSource=implementationSource, methodCodes=methodCodes)
        logger.debug("Method index %r", self.__methodD.items())

    def __getMethodInfo(self, implementationSource="reference", methodCodes=None):
        """ Get method implementation with the input implementation source.

        """
        methodCodes = methodCodes if methodCodes else ["calculation"]
        methodD = {}
        try:
            methodIndex = self.__dApi.getMethodIndex()
            for _, mrL in methodIndex.items():
                for mr in mrL:
                    mId = mr.getId()
                    catName = mr.getCategoryName()
                    atName = mr.getAttributeName()
                    mType = mr.getType()
                    if (catName, atName) not in methodD:
                        methodD[(catName, atName)] = []
                    methDef = self.__dApi.getMethod(mId)
                    logger.debug("Category %s attribute %s mId %r type %r methDef %r", catName, atName, mId, mType, methDef)
                    mSource = methDef.getImplementationSource()
                    mCode = methDef.getCode()
                    if mSource == implementationSource and mCode in methodCodes:
                        mPriority = methDef.getPriority()
                        mLang = methDef.getLanguage()
                        mImplement = methDef.getImplementation()
                        dD = {"METHOD_LANGUAGE": mLang, "METHOD_IMPLEMENT": mImplement, "METHOD_TYPE": mType, "METHOD_CODE": mCode, "METHOD_PRIORITY": mPriority}
                        methodD[(catName, atName)].append(dD)
                #
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        ##
        logger.debug("Method dictionary %r", methodD)
        return methodD

    def __invokeAttributeMethod(self, methodPath, dataContainer, catName, atName, **kwargs):
        """ Invoke the input attribute method
        """
        ok = False
        try:
            modulePath, methodName = self.__methodPathSplit(methodPath)
            mObj = self.__getModuleInstance(modulePath, **kwargs)
            theMeth = getattr(mObj, methodName, None)
            ok = theMeth(dataContainer, catName, atName, **kwargs)
        except Exception as e:
            logger.exception("Failed invoking attribute %s %s method %r with %s", catName, atName, methodPath, str(e))
        return ok

    def __invokeCategoryMethod(self, methodPath, dataContainer, catName, **kwargs):
        """ Invoke the input category method
        """
        ok = False
        try:
            modulePath, methodName = self.__methodPathSplit(methodPath)
            mObj = self.__getModuleInstance(modulePath, **kwargs)
            theMeth = getattr(mObj, methodName, None)
            ok = theMeth(dataContainer, catName, **kwargs)
        except Exception as e:
            logger.exception("Failed invoking category %s method %r with %s", catName, methodPath, str(e))
        return ok

    def __invokeDatablockMethod(self, methodPath, dataContainer, blockName, **kwargs):
        """ Invoke the input data block method
        """
        ok = False
        try:
            modulePath, methodName = self.__methodPathSplit(methodPath)
            mObj = self.__getModuleInstance(modulePath, **kwargs)
            theMeth = getattr(mObj, methodName, None)
            ok = theMeth(dataContainer, blockName, **kwargs)
        except Exception as e:
            logger.exception("Failed invoking block %s method %r with %s", blockName, methodPath, str(e))
        return ok

    def apply(self, dataContainer):
        """ Apply category, attribute and block level dictionary methods on the input data container.
        """
        kwargs = self.__kwargs
        mTupL = self.__getCategoryMethods()
        logger.debug("Category methods %r", mTupL)
        for catName, _, methodPath, _ in mTupL:
            self.__invokeCategoryMethod(methodPath, dataContainer, catName, **kwargs)

        mTupL = self.__getAttributeMethods()
        logger.debug("Attribute methods %r", mTupL)
        for catName, atName, methodPath, _ in mTupL:
            self.__invokeAttributeMethod(methodPath, dataContainer, catName, atName, **kwargs)

        mTupL = self.__getDatablockMethods()
        logger.debug("Datablock methods %r", mTupL)
        for blockName, _, methodPath, _ in mTupL:
            self.__invokeDatablockMethod(methodPath, dataContainer, blockName, **kwargs)

        return True

    def __getDatablockMethods(self):
        mL = []
        try:
            for (dictName, _), mDL in self.__methodD.items():
                for mD in mDL:
                    if mD["METHOD_TYPE"].lower() == "datablock":
                        methodPath = mD["METHOD_IMPLEMENT"]
                        mL.append((dictName, None, methodPath, mD["METHOD_PRIORITY"]))
            mL = sorted(mL, key=itemgetter(3))
            return mL
        except Exception as e:
            logger.exception("Failing dictName %s with %s", dictName, str(e))
        return mL

    def __getCategoryMethods(self):
        mL = []
        try:
            for (catName, _), mDL in self.__methodD.items():
                for mD in mDL:
                    if mD["METHOD_TYPE"].lower() == "category":
                        methodPath = mD["METHOD_IMPLEMENT"]
                        mL.append((catName, None, methodPath, mD["METHOD_PRIORITY"]))
            mL = sorted(mL, key=itemgetter(3))
            return mL
        except Exception as e:
            logger.exception("Failing catName %r with %s", catName, str(e))
        return mL

    def __getAttributeMethods(self):
        mL = []
        try:
            for (catName, atName), mDL in self.__methodD.items():
                for mD in mDL:
                    if mD["METHOD_TYPE"].lower() == "attribute":
                        methodPath = mD["METHOD_IMPLEMENT"]
                        mL.append((catName, atName, methodPath, mD["METHOD_PRIORITY"]))
            mL = sorted(mL, key=itemgetter(3))
            return mL
        except Exception as e:
            logger.exception("Failing catName %s atName %s with %s", catName, atName, str(e))
        return mL

    def __methodPathSplit(self, methodPath):
        """Extract module path and the method name from the input path.  Optional
           remap the module path.

        Arguments:
            methodPath {str} -- implementation path from dictionary method definition

        Returns:
            {tuple str} -- module path, method name
        """
        try:
            # Strip off any leading path of the module from the method path.
            mpL = str(methodPath).split(".")
            methodName = mpL[-1]
            tp = ".".join(mpL[:-1])
            modulePath = self.__modulePathMap[tp] if tp in self.__modulePathMap else tp
            return modulePath, methodName
        except Exception as e:
            logger.error("Failing for method path %r with %s", methodPath, str(e))
        return None, None

    def __getModuleInstance(self, modulePath, **kwargs):
        #
        if self.__cacheModuleFlag and modulePath in self.__moduleCache:
            return self.__moduleCache[modulePath]
        #
        mObj = None
        try:
            aMod = __import__(modulePath, globals(), locals(), [""])
            sys.modules[modulePath] = aMod
            #
            # Strip off any leading path to the module before we instaniate the module object.
            mpL = str(modulePath).split(".")
            moduleName = mpL[-1]
            #
            mObj = getattr(aMod, moduleName)(**kwargs)
            self.__moduleCache[modulePath] = mObj

        except Exception as e:
            logger.error("Failing to instance helper %r with %s", modulePath, str(e))
        return mObj



##
# File:  MyDdAdapter.py
# Date:  10-April-2014  J.Westbrook
#
# Updates:
#
# 11-April-2014  jdw Generalized from WFTaskRequestDBAdapter.py
# 13-April-2014  jdw working with workflow schema WFTaskRequest() -
# 19-Feb  -2015  jdw various fixes
# 10-July -2015  jdw Change method/class names from MySqlGen
# 10-March-2018  jdw  Py2->Py3 compatibility using driver fork described at https://mysqlclient.readthedocs.io/user_guide.html#
# 29-March-2018  jdw remove dependency on wwPDB configuration  -   Use generic configuratio object in constructor -
#  9-July -2018  jdw flip back to time.time()
#
###
##
"""
Database adapter for managing simple access and persistance queries using a MySQL relational database store.
"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"


import copy
import logging
import time

from rcsb.db.mysql.Connection import Connection
from rcsb.db.mysql.MyDbUtil import MyDbQuery
from rcsb.db.sql.SqlGen import SqlGenAdmin, SqlGenCondition, SqlGenQuery

logger = logging.getLogger(__name__)


class MyDbAdapter(object):

    """ Database adapter for managing simple access and persistance queries using a MySQL relational database store.
    """

    def __init__(self, schemaDefObj, cfgOb=None, verbose=False):
        self.__verbose = verbose
        self.__debug = False
        self.__cfgOb = cfgOb
        #
        self.__sD = schemaDefObj
        self.__databaseName = self.__sD.getDatabaseName()
        self.__dbCon = None
        self.__cObj = None
        self.__defaultD = {}
        self.__attributeParameterMap = {}
        self.__attributeConstraintParameterMap = {}

    def __open(self, infoD):
        cObj = Connection()
        cObj.setPreferences(infoD)
        ok = cObj.openConnection()
        if ok:
            return cObj
        else:
            return None

    def __close(self, cObj):
        if cObj is not None:
            cObj.closeConnection()
            return True
        else:
            return False

    def __getClientConnection(self, cObj):
        return cObj.getClientConnection()

    def _open(self, dbServer=None, dbHost=None, dbName=None, dbUser=None, dbPw=None, dbSocket=None, dbPort=None):
        """  Open a connection to the data base server hosting WF status and tracking data -

             Internal configuration details will be used if these are not externally supplied.
        """
        infoD = {}
        infoD["DB_HOST"] = dbHost if dbHost is not None else self.__cfgOb.get("SITE_DB_HOST_NAME")
        infoD["DB_PORT"] = dbPort if dbPort is not None else self.__cfgOb.get("SITE_DB_PORT_NUMBER")
        infoD["DB_NAME"] = dbName if dbName is not None else self.__cfgOb.get("SITE_DB_DATABASE_NAME")
        infoD["DB_USER"] = dbUser if dbUser is not None else self.__cfgOb.get("SITE_DB_USER_NAME")
        infoD["DB_PW"] = dbPw if dbPw is not None else self.__cfgOb.get("SITE_DB_PASSWORD")
        infoD["DB_SERVER"] = dbServer if dbServer is not None else self.__cfgOb.get("SITE_DB_SERVER")
        infoD["DB_SOCKET"] = dbSocket if dbSocket is not None else self.__cfgOb.get("SITE_DB_SOCKET")
        #
        self.__cObj = self.__open(infoD)
        self.__dbCon = self.__getClientConnection(self.__cObj)
        return self.__dbCon is not None

    def _close(self):
        """  Close connection to the data base server hosting WF status and tracking data -
        """
        if self.__dbCon is not None:
            self.__close(self.__cObj)
            self.__dbCon = None
            self.__cObj = None

    def _setDebug(self, flag=True):
        self.__debug = flag

    def _setDataStore(self, dataStoreName):
        """  Set/reassign the database for all subsequent transactions.
        """
        self.__databaseName = dataStoreName

    def _getParameterDefaultValues(self, contextId):
        if contextId is not None and contextId in self.__defaultD:
            return self.__defaultD[contextId]
        else:
            return {}

    def _setParameterDefaultValues(self, contextId, valueD):
        """  Set the optional lookup dictionary of default values for unspecified parameters...

             valueD = { 'paramName1': <default value1>,  'paramName2' : <default value2>, ...  }
        """
        self.__defaultD[contextId] = copy.deepcopy(valueD)
        return True

    def _setAttributeParameterMap(self, tableId, mapL):
        """  Set list of correspondences between method parameters and table attribute IDs.

             These correspondences are used to map key-value parameter pairs to their associated table attribute values.

             mapL=[ (atId1,paramName1),(atId2,paramName2),... ]
        """
        self.__attributeParameterMap[tableId] = mapL
        return True

    def _getDefaultAttributeParameterMap(self, tableId):
        """ Return default attributeId parameter name mappings for the input tableId.

             mapL=[ (atId1,paramName1),(atId2,paramName2),... ]
        """
        return self.__sD.getDefaultAttributeParameterMap(tableId)

    def _getAttributeParameterMap(self, tableId):
        """
           For the input table return the method keyword argument name to table attribute mapping -
        """
        if tableId is not None and tableId in self.__attributeParameterMap:
            return self.__attributeParameterMap[tableId]
        else:
            return []

    def _getConstraintParameterMap(self, tableId):
        """
           For the input table return the method keyword argument name to table attribute mapping for
           those attributes that serve as constraints for update transactions -

        """
        if tableId is not None and tableId in self.__attributeConstraintParameterMap:
            return self.__attributeConstraintParameterMap[tableId]
        else:
            return []

    def _setConstraintParameterMap(self, tableId, mapL):
        """  Set list of correspondences between method parameters and table attribute IDs to be used as
             contraints in update operations.

             These correspondences are used to map key-value paramter pairs to their associated table attribute values.

             mapL=[ (atId1,paramName1),(atId2,paramName2),... ]
        """
        self.__attributeConstraintParameterMap[tableId] = mapL
        return True

    def _createSchema(self):
        """ Create table schema using the current class schema definition
        """
        if self.__debug:
            startTime = time.time()
            logger.info("Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        ret = False
        try:
            iOpened = False
            if self.__dbCon is None:
                self._open()
                iOpened = True
            #
            tableIdList = self.__sD.getSchemaIdList()
            myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
            myAd = SqlGenAdmin(self.__verbose)

            for tableId in tableIdList:
                sqlL = []
                tableDefObj = self.__sD.getSchemaObject(tableId)
                sqlL.extend(myAd.createTableSQL(databaseName=self.__databaseName, tableDefObj=tableDefObj))

                ret = myQ.sqlCommand(sqlCommandList=sqlL)
                if self.__verbose:
                    logger.info("For tableId %s server returns: %s\n", tableId, ret)
                if self.__debug:
                    logger.info("SQL: %s\n", "\n".join(sqlL))
            if iOpened:
                self._close()
        except Exception as e:
            status = " table create error " + str(e)
            logger.info("status %s\n", status)
            if self.__verbose:
                logger.exception("Failing with %s", str(e))

        if self.__debug:
            endTime = time.time()
            logger.info("Completed at %s (%.3f seconds)\n", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ret

    def _getSecondsSinceEpoch(self):
        """  Return number of seconds since the epoch at the precision of the local installation.
             Typically a floating point value with microsecond precision.

             This is used as the default time reference (e.g. timestamp) for monitoring task requests.
        """
        return time.time()

    def _insertRequest(self, tableId, contextId, **kwargs):
        """ Insert into the input table using the keyword value pairs provided as input arguments -

            The contextId controls the handling default values for unspecified parameters.
        """
        startTime = time.time()
        if self.__debug:
            logger.info("Starting at %s\n", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        ret = False
        try:
            iOpened = False
            if self.__dbCon is None:
                self._open()
                iOpened = True
            #
            tableDefObj = self.__sD.getSchemaObject(tableId)
            #
            myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
            myAd = SqlGenAdmin(self.__verbose)
            defaultValD = self._getParameterDefaultValues(contextId=contextId)
            #
            # Create the attribute and value list for template --
            #
            vList = []
            aList = []
            for atId, kwId in self._getAttributeParameterMap(tableId=tableId):
                if kwId in kwargs and kwargs[kwId] is not None:
                    vList.append(kwargs[kwId])
                    aList.append(atId)
                else:
                    # use the default values if these exist
                    if kwId in defaultValD and defaultValD[kwId] is not None:
                        vList.append(defaultValD[kwId])
                        aList.append(atId)
                    else:
                        # appropriate null handling -- all fields must be assigned on insert --
                        vList.append(tableDefObj.getAppNullValue(atId))
                        aList.append(atId)

            sqlT = myAd.idInsertTemplateSQL(self.__databaseName, tableDefObj, insertAttributeIdList=aList)
            if self.__debug:
                logger.info("aList %d vList %d\n", len(aList), len(vList))
                logger.info("insert template sql=\n%s\n", sqlT)
                logger.info("insert values vList=\n%r\n", vList)
                # sqlC = sqlT % vList
                # logger.info("insert sql command =\n%s\n",  sqlC)
            ret = myQ.sqlTemplateCommand(sqlTemplate=sqlT, valueList=vList)
            if iOpened:
                self._close()

        except Exception as e:
            status = " insert operation error " + str(e)
            logger.info("status %s\n", status)
            if self.__verbose:
                logger.exception("Failing with %s", str(e))
        if self.__debug:
            endTime = time.time()
            logger.info("Completed %s (%.3f seconds)\n", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)

        return ret

    def _updateRequest(self, tableId, contextId, **kwargs):
        """  Update the input table using the keyword value pairs provided as input arguments -

             The contextId controls the handling default values for unspecified parameters.

        """
        startTime = time.time()
        if self.__debug:
            logger.info("Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        ret = False
        try:
            iOpened = False
            if self.__dbCon is None:
                self._open()
                iOpened = True
            #
            tableDefObj = self.__sD.getSchemaObject(tableId)
            #
            myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
            myAd = SqlGenAdmin(self.__verbose)
            defaultValD = self._getParameterDefaultValues(contextId=contextId)
            cIdList = self._getConstraintParameterMap(tableId)

            #
            # create the value list for template --
            #
            vList = []
            aList = []
            cList = []
            for atId, kwId in self._getAttributeParameterMap(tableId):
                if (atId, kwId) in cIdList:
                    continue
                if kwId in kwargs and kwargs[kwId] is not None:
                    vList.append(kwargs[kwId])
                    aList.append(atId)
                else:
                    if kwId in defaultValD and defaultValD[kwId] is not None:
                        vList.append(defaultValD[kwId])
                        aList.append(atId)

            for atId, kwId in cIdList:
                if kwId in kwargs and kwargs[kwId] is not None:
                    vList.append(kwargs[kwId])
                    cList.append(atId)

            sqlT = myAd.idUpdateTemplateSQL(self.__databaseName, tableDefObj, updateAttributeIdList=aList, conditionAttributeIdList=cList)
            if self.__debug:
                logger.info("update sql: %s", sqlT)
                logger.info("update values: %r", vList)
            ret = myQ.sqlTemplateCommand(sqlTemplate=sqlT, valueList=vList)
            if iOpened:
                self._close()

        except Exception as e:
            status = " update operation error " + str(e)
            logger.info("status %s", status)
            if self.__verbose:
                logger.exception("Failing with %s", str(e))
        if self.__debug:
            endTime = time.time()
            logger.info("Completed at %s (%.3f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ret

    def _select(self, tableId, **kwargs):
        """ Construct a selection query for input table and optional constraints provided as keyword value pairs in the
            input arguments.  Return a list of dictionaries of these query details including all table attributes.
        """
        startTime = time.time()
        if self.__debug:
            logger.info("Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        rdList = []
        try:
            iOpened = False
            if self.__dbCon is None:
                self._open()
                iOpened = True
            #
            tableDefObj = self.__sD.getSchemaObject(tableId)
            myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
            sqlGen = SqlGenQuery(schemaDefObj=self.__sD, verbose=self.__verbose)
            sqlGen.setDatabase(databaseName=self.__databaseName)
            sqlConstraint = SqlGenCondition(schemaDefObj=self.__sD, verbose=self.__verbose)
            #
            atMapL = self._getAttributeParameterMap(tableId=tableId)
            for kwArg, _ in kwargs.items():
                for atId, kwId in atMapL:
                    if kwId == kwArg:
                        if tableDefObj.isAttributeStringType(atId):
                            cTup = ((tableId, atId), "EQ", (kwargs[kwId], "CHAR"))
                        else:
                            cTup = ((tableId, atId), "EQ", (kwargs[kwId], "OTHER"))
                        sqlConstraint.addValueCondition(cTup[0], cTup[1], cTup[2])
                        break
            #
            # Add optional constraints OR ordering by primary key attributes
            if sqlConstraint.get():
                sqlGen.setCondition(sqlConstraint)
            else:
                for atId in tableDefObj.getPrimaryKeyAttributeIdList():
                    sqlGen.addOrderByAttributeId(attributeTuple=(tableId, atId))

            atIdList = self.__sD.getAttributeIdList(tableId)
            for atId in atIdList:
                sqlGen.addSelectAttributeId(attributeTuple=(tableId, atId))
            #
            sqlS = sqlGen.getSql()
            if self.__debug:
                logger.info("selection sql: %s", sqlS)

            rowList = myQ.selectRows(queryString=sqlS)
            sqlGen.clear()
            #
            # return the result set as a list of dictionaries
            #
            for iRow, row in enumerate(rowList):
                rD = {}
                for colVal, atId in zip(row, atIdList):
                    rD[atId] = colVal
                if self.__debug:
                    logger.info("result set row %d dictionary %r", iRow, rD.items())
                rdList.append(rD)
            if iOpened:
                self._close()
        except Exception as e:
            status = " operation error " + str(e)
            logger.info("status %s", status)
            if self.__verbose:
                logger.exception("Failing with %s", str(e))

        if self.__debug:
            endTime = time.time()
            logger.info("Completed at %s (%.3f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return rdList

    def _deleteRequest(self, tableId, **kwargs):
        """ Delete from input table records identified by the keyword value pairs provided as input arguments -
        """
        startTime = time.time()
        if self.__debug:
            logger.info("Starting at %s", time.strftime("%Y %m %d %H:%M:%S", time.localtime()))
        ret = False
        try:
            iOpened = False
            if self.__dbCon is None:
                self._open()
                iOpened = True

            tableDefObj = self.__sD.getSchemaObject(tableId)
            #
            #
            myQ = MyDbQuery(dbcon=self.__dbCon, verbose=self.__verbose)
            myAd = SqlGenAdmin(self.__verbose)
            #
            # Create the attribute and value list for template --
            #
            vList = []
            aList = []
            for atId, kwId in self._getAttributeParameterMap(tableId):
                if kwId in kwargs and kwargs[kwId] is not None:
                    vList.append(kwargs[kwId])
                    aList.append(atId)

            sqlT = myAd.idDeleteTemplateSQL(self.__databaseName, tableDefObj, conditionAttributeIdList=aList)
            if self.__debug:
                logger.info("delete sql: %s", sqlT)
                logger.info("delete values: %r", vList)
            ret = myQ.sqlTemplateCommand(sqlTemplate=sqlT, valueList=vList)

            if iOpened:
                self._close()

        except Exception as e:
            status = " delete operation error " + str(e)
            logger.info("status %s", status)
            if self.__verbose:
                logger.exception("Failing with %s", str(e))

        if self.__debug:
            endTime = time.time()
            logger.info("Completedat %s (%.3f seconds)", time.strftime("%Y %m %d %H:%M:%S", time.localtime()), endTime - startTime)
        return ret
